package jwars;

import java.io.IOException;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.URL;
import java.util.Enumeration;
import javax.imageio.ImageIO;

import java.awt.Graphics2D;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Transparency;
import java.awt.image.BufferedImage;

import jwars.model.WarModel;
import jwars.synch.Timer;

/**
 * This class contains miscellaneous static data and methods useful
 * throughout JWars.
 */
public final class Util{

  //Basic program information
  public final static String
      GAME_NAME = "JWars",
      FULL_GAME_NAME = "JWars Warfare Simulation",
      VERSION = "alpha 2 build 6",
      NAME_AND_VERSION = GAME_NAME + " " + VERSION,
      AUTHORS = "Michael Francker Christensen & Ask Hjorth Larsen";

  //Mimics System.out, though we sadly cannot use final modifier
  public static PrintWriter out = new PrintWriter(System.out);

  //The Timer used for fixing the framerate of JWars
  public final static Timer TIMER = new Timer(WarModel.TIME_PER_FRAME);

  /**
   * Redirect the JWars printing facility
   * @param out PrintWriter
   */
  public static void setOut(PrintWriter out){
    Util.out = out;
  }

  //Port used for networking
  public final static int PORT = 7777;

  //Display settings
  public final static GraphicsDevice GRAPH_DEVICE = GraphicsEnvironment.
      getLocalGraphicsEnvironment().getDefaultScreenDevice();

  public final static GraphicsConfiguration GRAPH_CONFIG =
      GRAPH_DEVICE.getDefaultConfiguration();

  static boolean isFullScreen = false;

  public final static boolean isFullScreen(){
    return isFullScreen;
  }

  private final static int
      DEFAULT_WINDOWED_DISPLAY_SIZE_X = 800,
      DEFAULT_WINDOWED_DISPLAY_SIZE_Y = 600;

  public final static int DISPLAY_SIZE_X(){
    return isFullScreen() ?
        GRAPH_DEVICE.getDisplayMode().getWidth() :
        DEFAULT_WINDOWED_DISPLAY_SIZE_X;
  }

  public final static int DISPLAY_SIZE_Y(){
    return isFullScreen() ?
        GRAPH_DEVICE.getDisplayMode().getHeight() :
        DEFAULT_WINDOWED_DISPLAY_SIZE_Y;
  }

  private static String userName = "Manstein";

  public static void setUserName(String userName){
    Util.userName = userName;
  }

  public static String getUserName(){
    return userName;
  }

  /**
   * Creates a hopefully managed, opaque image of specified dimensions
   * @param width int
   * @param height int
   * @return BufferedImage
   */
  public static BufferedImage createImage(int width, int height){
    return GRAPH_CONFIG.createCompatibleImage(width, height,
        Transparency.OPAQUE);
  }

  /**
   * Creates a hopefully managed image of specified dimensions and transparency.
   * @param width int
   * @param height int
   * @param transparency int
   * @return BufferedImage
   */
  public static BufferedImage createImage(int width, int height,
      int transparency){
    return GRAPH_CONFIG.createCompatibleImage(width, height, transparency);
  }

  /**
   * Load a bitmasked image from the file denoted by argument String,
   * relative to the top-level JWars directory
   * @param path String
   * @return BufferedImage
   */
  public static BufferedImage readImage(String path){
    return readImage(path, Transparency.BITMASK);
  }

  /**
   * Loads an image with specified transparency from the file denoted by
   * argument String, relative to the top-level JWars directory
   * @param path String
   * @param transparency int
   * @return BufferedImage
   */
  public static BufferedImage readImage(String path, int transparency){
    BufferedImage src = null;
    BufferedImage dst = null;
    String resource = "/graphics/" + path;
    URL url = new Object().getClass().getResource(resource);
    if(url == null){
      System.err.println("<Util> ***WARNING***");
      System.err.println("<Util> Image resource \"" + resource + "\" could not be loaded. You probably forgot to include the graphics directory... AGAIN!");
      System.err.println("<Util> Returning EMPTY IMAGE instead.");
      return createImage(1, 1, Transparency.BITMASK);
    }
    try{
      src = ImageIO.read(url);
    } catch(IOException ex){
      ex.printStackTrace();
    }
    dst = createImage(src.getWidth(), src.getHeight(), transparency);
    Graphics2D g = dst.createGraphics();
    g.drawImage(src, 0, 0, null);
    g.dispose();
    return dst;
  }

  /**
   * Creates a new, hopefully managed BufferedImage identical to the argument
   * (possibly not managed) BufferedImage. Images that have had their data
   * modified directly are generally not managed.
   * @param image Image
   * @return BufferedImage
   */
  public static BufferedImage recreate(BufferedImage image){
    BufferedImage dst = createImage(image.getWidth(null), image.getHeight(null),
        image.getTransparency());
    Graphics2D g2d = dst.createGraphics();
    g2d.drawImage(image, 0, 0, null);
    g2d.dispose();
    return dst;
  }

  public static double minusPiToPi(double angle){
    double pi2 = 2 * Math.PI;
    if(angle >= Math.PI || angle <= -Math.PI){
      angle = angle % (pi2);
      if(angle < -Math.PI){
        angle += pi2;
      }
      if(angle > Math.PI){
        angle -= pi2;
      }
    }
    return angle;
  }

  public static String getLocalIPAddress(){
    try{
      Enumeration interfaces = NetworkInterface.getNetworkInterfaces();
      while(interfaces.hasMoreElements()){
        NetworkInterface ni = (NetworkInterface)interfaces.nextElement();
        Enumeration addresses = ni.getInetAddresses();
        while(addresses.hasMoreElements()){
          InetAddress ip = (InetAddress)addresses.nextElement();
          //No 127.0.0.1 and no MAC addresses
          if(!ip.isLoopbackAddress() && ip.getHostAddress().indexOf(":") == -1){
            return ip.getHostAddress();
          }
        }
      }
    } catch(SocketException ex){
      ex.printStackTrace();
    }
    return "";
  }

  final static int MEANING_OF_LIFE_THE_UNIVERSE_AND_EVERYTHING = 42;

  private Util(){}

  static{
    jwars.test.explosions.Explosions.init();
  }

}
