package jwars;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import jwars.control.Community;
import jwars.control.InputManager;
import jwars.data.Loader;
import jwars.event.QueueManager;
import jwars.gui.Console;
import jwars.gui.WarFrame;
import jwars.gui.WarPanel;
import jwars.model.Moveable;
import jwars.model.WarModel;
import jwars.model.units.Teams;
import jwars.net.Client;
import jwars.net.IOHandler;
import jwars.net.JWarIOHandler;
import jwars.synch.Timer;
import jwars.synch.TimerListener;

/**
 * This class represents a complete game of JWars, managing the initialization
 * and updating of graphics as well as logic separately.
 *
 * This class is used only clientside. A server must be created separately,
 * see the class WarServer.
 */
public class WarSession
    implements TimerListener{

  private static SynchInstruction synchInstruction = new SynchInstruction();
  private LaunchInstruction launcher = new LaunchInstruction();

  private WarFrame wFrame;
  private InputManager wControl;
  private Timer timer;
  private WarPanel wPanel;
  private QueueManager queues;
  private Client client;
  private WarMenu menu;

  private int lastGraphicsUpdate = 0;
  private int ensureGraphicsUpdateInterval = 100;

  /**
   * Create a session of JWars, connecting to the server specified by ip and
   * port.
   * @param ip String representation of the server IP,
   * on the form ###.###.###.###
   * @param port Port number of the connection to the server.
   */
  public WarSession(WarMenu menu, String ip, int port){
    WarModel.WORLD.init();
    loadUnits();
    queues = WarModel.QUEUES;
    queues.register(Moveable.getUnitMoveQueue());
    wControl = new InputManager(queues.getInstantQueue());
    Console console = new Console(wControl);

    wFrame = new WarFrame(wControl);
    wPanel = wFrame.getWarPanel();

    Util.setOut(console.getPrintWriter());
    wControl.setOut(console.getPrintWriter());
    menu.expandInterface(console);
    Util.out.println("Welcome to " + Util.NAME_AND_VERSION +
                     ". Use the text field below to type messages.");

    timer = Util.TIMER;
    timer.addTimerListener(this);
    client = new Client(JWarIOHandler.PROTOCOL, Community.players);
    Community.useClient(client);
    this.menu = menu;
    try{
      client.connect(ip, port);
    } catch(Exception ex){
      javax.swing.JOptionPane.showMessageDialog(null, "Could not connect to server!");
      System.exit(0);
    }
    //boolean isConnected = client.connect(ip, port);
    //if(!isConnected)
    //  System.out.println("Could not connect to server");
  }

  /**
   * Update the complete state of this session of JWars. Will update the logic
   * and, if there is time, graphics.
   * @param frameCount int
   * @param isLate boolean
   */
  public void update(int frameCount, boolean isLate){
    if(stasisBadnessCounter > 0){
      stasisBadnessCounter -= 1;
    }
    logicUpdate(frameCount);
    if(!isLate || (frameCount - lastGraphicsUpdate)*WarModel.TIME_PER_FRAME > ensureGraphicsUpdateInterval){
      graphicsUpdate(frameCount);
    }
    if(frameCount == synchInstruction.maxTime - 1){
      //Not "== maxTime", because then other instructions may arrive in the
      //same frame and be executed late.
      //System.out.println("<WarSession> Waiting for host at frame number "+frameCount);
      timer.suspend();
    } else if(frameCount >= synchInstruction.maxTime){
      throw new RuntimeException(
          "<WarSession> Out of synch! Max time: "+synchInstruction.maxTime+"; Current time: "+
          frameCount);
    }
  }

  /**
   * Update the logical state of this session of JWars.
   * @param time current frame number
   */
  private void logicUpdate(int time){
    queues.update(time);
    WarModel.WORLD.update();
  }

  /**
   * Perform all graphical updates for the current frame.
   * @param time current frame number
   */
  private void graphicsUpdate(int time){
    this.lastGraphicsUpdate = time;
    if(time % (1000/WarModel.TIME_PER_FRAME) == 0){
      wFrame.drawPanels();
    }
    wPanel.render();
  }

  private int stasisBadnessCounter = 0;

  /**
   * React to the game being paused or suspended; see the Timer class for
   * details. This method currently does nothing.
   */
  public void reportStasis(){
    int time = timer.getTime();
    /**
     * The following will attempt to stabilize game play if the timer is
     * suspended too frequently
     */
    /*if(timer.isSuspended()){
      stasisBadnessCounter += 10;
      if(stasisBadnessCounter > 100){
        stasisBadnessCounter = 0;
        timer.delay(1000);
        System.out.println("Delaying timer");
      }
    }*/
    graphicsUpdate(time);
    if(synchInstruction.maxTime > time+1){
      //System.out.println("<WarSession> Unpausing Timer; current time: "+timer.getTime());
      timer.unpause();
    }
  }

  /**
   * Send the instruction to launch the game. This method is called from a
   * client (presently the client who has created the server).
   */
  public void sendLaunchInstruction(){
    try{
      launcher.send(null);
    } catch(IOException ex){
      ex.printStackTrace();
    }
  }

  private void loadUnits(){
    Loader loader = new Loader(WarModel.WORLD.meter, WarModel.WORLD.second);
    System.out.println("Loading game data...");
    loader.load("/data/data.dat");
    System.out.println("All data loaded.");
    System.out.println(""+loader);
    new Teams(WarModel.WORLD, loader.getUnitTable());
  }

  /**
   * Represents the instruction to launch the actual game.
   */
  private class LaunchInstruction extends JWarIOHandler<Object>{

    protected void write(Object t, DataOutput output) throws IOException{
      System.out.println("Sending Launch Instruction!");
    }

    protected void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException{
      System.out.println("Launch Instruction echoed by server");
    }

    protected void read(DataInput input, int timeStamp) throws IOException{
      System.out.println("Launch Instruction received. Starting Timer!");
      menu.setVisible(false);
      Console c = wFrame.getConsole();
      Util.setOut(c.getPrintWriter());
      wControl.setOut(c.getPrintWriter());
      wFrame.setVisible(true);
      menu.dispose();
      menu = null;
      timer.start();
    }

  }

  public static IOHandler getDefaultSynchInstruction(){
    return synchInstruction;
  }

  /**
   * Represents the default instruction which the server sends across
   * the network with regular intervals. On receiving this instruction,
   * the game will be allowed to continue until such time as the frame count
   * reaches the time stamp of this instruction.
   */
  private static class SynchInstruction extends IOHandler{

    int maxTime = ServerSession.LATENCY;

    protected void read(DataInput input, int timeStamp) throws IOException{
      //System.out.println("<WarSession> Receiving synch instruction");
      maxTime = timeStamp;
    }

    protected void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException{
      //System.out.println("<WarSession> Sending synch instruction");
    }

    protected void write(Object t, DataOutput output) throws IOException{
    }

  }

}
