package jwars.control;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;

import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import jwars.Util;
import jwars.event.Event;
import jwars.event.EventQueue;
import jwars.event.SynchronizedQueue;
import jwars.gui.TextParser;
import jwars.model.Unit;
import jwars.model.WarModel;
import jwars.net.JWarIOHandler;
import jwars.test.Player;

/**
 * Contains a collection of commands, each of which may be invoked
 * by means of a supplying its identifying String or, if a key is bound
 * to that command, by pressing the appropriate key when a window which uses
 * this WarController as a KeyListener has focus.
 *
 * That component should add this WarController as FocusListener as well.
 * This is to ensure that commands which are meant to be repeated as long as the
 * corresponding key is down will stop repeating when the component loses
 * focus.
 */
public class InputManager
    implements KeyListener, FocusListener, TextParser{

  //Any input String to the parse method starting with this character
  //is considered a command.
  private final char COMMAND_CHAR = '/';

  //Strings passed to the parse method which are not considered commands
  //will be written to this PrintWriter if it is non-null.
  private PrintWriter out;

  //Used for sending Strings across a network
  private StringIOHandler stringSerializer = new StringIOHandler();

  //private final QueueManager qm;
  public final SynchronizedQueue DEFAULT_QUEUE;
  /**
   * The keycodes corresponding to standard keyboard input are all between
   * 0 and 255, and thus we can let them be in an array of length 256
   */
  private KeyTrackEvent[] keyCommands = new KeyTrackEvent[0xFF];

  /**
   * Dictionary of registered events.
   */
  private HashMap<String, Event> consoleCommands = new HashMap<String, Event>();

  /**
   * Create a WarController using the specified queue manager.
   * @param qm QueueManager
   */
  public InputManager(SynchronizedQueue instantQueue){
    //this.instantQueue = instantQueue;
    this.DEFAULT_QUEUE = instantQueue;
    //Bind all keys to do nothing (so they don't throw nullpointerexceptions)
    Arrays.fill(keyCommands, new EmptyEvent());
    put("currentthread", new Event(DEFAULT_QUEUE){
      protected void execute(){
        Util.out.println(Thread.currentThread());
      }

    });
    put("countunits", new Event(DEFAULT_QUEUE){
      protected void execute(){
        Unit selection = Control.UNIT_SELECTION.markedUnit();
        if(selection == null){
          Util.out.println("Total units in game: " +
                           WarModel.WORLD.countTotalUnits());
        } else {
          Util.out.println("Total units in selection "+WarModel.WORLD.countSubUnits(selection));
        }
      }

    });
    put("commands", new Event(DEFAULT_QUEUE){
      protected void execute(){
        Iterator<String> keys = consoleCommands.keySet().iterator();
        Util.out.println("---Commands---");
        while(keys.hasNext()){
          Util.out.println(""+keys.next());
        }
      }
    });
    put("lateness", new Event(DEFAULT_QUEUE){
      protected void execute(){
        Util.out.println("Milliseconds late:"+(-Util.TIMER.timeLeftMillis()));
      }
    });
    put("stop", KeyEvent.VK_S, new KeyTrackEvent(DEFAULT_QUEUE){
      protected void execute(){
        Unit markedUnit = Control.UNIT_SELECTION.markedUnit();
        if (markedUnit != null)
          markedUnit.sendStopCommand();
      }
    });
  }

  /**
   * Store and associate specified event with name and key code. The event will
   * be executed when either name or key code are received through
   * the parse and keyPressed methods of this WarController.
   * @param name String
   * @param keyCode int
   * @param event Event
   */
  public void put(String name, int keyCode, KeyTrackEvent event){
    this.keyCommands[keyCode] = event;
    put(name, event);
  }

  /**
   * Store and associate specified event with a name. The event will
   * be executed when the name is received through
   * the parse method of this WarController.
   * @param name String
   * @param event Event
   */
  public void put(String name, Event event){
    consoleCommands.put(name.toLowerCase(), event);
  }

  /**
   * Fetch the command associated with the key code of argument event and
   * schedule it for immediate execution
   * @param e KeyEvent
   */
  public void keyPressed(KeyEvent e){
    int keyCode = e.getKeyCode();
    if(keyCode < keyCommands.length && !keyCommands[keyCode].isKeyDown){ //Perhaps use keyCode & 0xFF
      keyCommands[keyCode].isKeyDown = true;
      keyCommands[keyCode].enqueue();
    }
  }

  public void keyTyped(KeyEvent e){}

  /**
   * Ensures that the command corresponding to the code of the argument KeyEvent
   * is no longer to be executed.
   * @param e KeyEvent
   */
  public void keyReleased(KeyEvent e){
    int keyCode = e.getKeyCode();
    if(keyCode < keyCommands.length){
      keyCommands[keyCode].isKeyDown = false;
    }
  }

  /**
   * Fetch  the command associated with the argument String and
   * schedule it for immediate execution, or, if no command is associated
   * with arguement String, pass the String to a StringSerializer.
   * @param s String
   */
  public String parse(String s){
    if(s.charAt(0) == COMMAND_CHAR){ //Execute command
      Event e = consoleCommands.get(s.substring(1, s.length()).toLowerCase());
      if(e != null){
        e.enqueue();
        return s;
      } else {
        return "Unrecognized command: "+s.substring(1, s.length());
      }
    } else { //Not a command
      try{
        stringSerializer.send(s);
      } catch(IOException ex){
        ex.printStackTrace();
        System.out.println("<WarController> Cannot send String");
      }
    }
    return null;
  }

  /**
   * When losing focus, release keys
   * @param focusEvent FocusEvent
   */
  public void focusLost(FocusEvent e){
    for(KeyTrackEvent event : keyCommands){
      event.isKeyDown = false;
    }
  }

  public void focusGained(FocusEvent e){}

  /**
   * Represents actions corresponding to certain keys being held down.
   * KeyTrackEvents will automatically continue
   */
  public abstract static class KeyTrackEvent
      extends Event{

    boolean isKeyDown = false;

    public KeyTrackEvent(EventQueue queue){
      super(queue);
    }

    /**
     * If this event should be executed continuously while the key is pressed,
     * extending classes should call super.execute() from this method when
     * overriding.
     */
    protected void execute(){
      if(isKeyDown){
        enqueue();
      }
    }

  }

  /**
   * Sets the PrintWriter to which this WarController passes text which is
   * received from a network. This funcionality may be moved to another class
   * at some time.
   * @param out PrintWriter
   */
  public void setOut(PrintWriter out){
    this.out = out;
  }

  /**
   * Used for unbound keys
   */
  private class EmptyEvent
      extends KeyTrackEvent{

    public EmptyEvent(){
      super(DEFAULT_QUEUE);
    }

    /* Do nothing*/
    protected void execute(){}

  }

  /**
   * Responsible for sending and receiving Strings across a network.
   */
  public class StringIOHandler extends JWarIOHandler<String>{

    protected void read(DataInput input, int timeStamp) throws IOException{
      int clientIdentifier = input.readInt();
      Player p = Community.players.getPlayer(clientIdentifier);
      String msg = input.readUTF();
      if(out != null){
        out.print("<"+p.getName()+"> ");
        out.println(msg);
      }
    }

    protected void echo(int clientIdentifier, DataInput input, DataOutput output) throws IOException{
      output.writeInt(clientIdentifier);
      output.writeUTF(input.readUTF());
    }

    protected void write(String s, DataOutput output) throws IOException{
      output.writeUTF(s);
    }

  }

}
