package jwars.gui;

import java.io.PrintWriter;
import java.io.Writer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTextPane;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyledDocument;

public class Console
    extends JPanel{

  private JTextPane outputPane = new JTextPane();
  private JTextField inputField = new JTextField();
  private StyledDocument log = outputPane.getStyledDocument();
  private Component defaultFocusedComponent = null;
  private PrintWriter out = new PrintWriter(new ConsoleWriter(), true);
  private TextParser parser;

  private JScrollPane scrollPane =
      new JScrollPane(outputPane, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                      JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

  private Style defaultStyle = outputPane.addStyle("",null);
  private Style inputStyle = outputPane.addStyle("",null);

  /**
   * Creates a console which does not use a TextParser.
   */
  public Console(){
    this(new TextParser(){
      public String parse(String s){
        return s;
      }
    });
  }

  /**
   * Create a Console which returns focus to the argument Component
   * after having been used. Input Strings will be passed to the argument
   * parser.
   *
   * @param defaultFocusedComponent Component
   */
  public Console(TextParser parser){
    this.parser = parser;
    this.setLayout(new BorderLayout());

    inputField.setFont(new Font("Monospaced", Font.PLAIN, 12));
    outputPane.setFont(new Font("Monospaced", Font.PLAIN, 12));

    outputPane.setEditable(false);

    outputPane.setRequestFocusEnabled(false);
    inputField.setRequestFocusEnabled(false);

    ConsoleListener listener = new ConsoleListener();
    inputField.addActionListener(listener);
    inputField.addMouseListener(listener);

    this.add(scrollPane, BorderLayout.CENTER);
    this.add(inputField, BorderLayout.SOUTH);

    generateStyles();
  }

  /**
   * Allow this Console focus such that it can listen for input
   */
  public void enableTyping(){
    inputField.setEnabled(true);
    inputField.requestFocusInWindow();
  }

  public void println(String text, Style style){
    print(text+'\n', style);
  }

  public void print(String text, Style style){
    try{
      log.insertString(log.getLength(), text, style);
    } catch(BadLocationException ex){
      ex.printStackTrace();
    }
    outputPane.setCaretPosition(log.getLength());
  }

  private void generateStyles(){
    StyleConstants.setForeground(defaultStyle, Color.black);
    StyleConstants.setForeground(inputStyle, Color.black);
  }

  public JTextPane getOutputPane(){
    return outputPane;
  }

  public JTextField getInputField(){
    return inputField;
  }

  /**
   * Get the PrintWriter associated with this Console
   * @return PrintWriter the PrintWriter
   */
  public PrintWriter getPrintWriter(){
    return out;
  }

  public void setDefaultFocusedComponent(Component comp){
    this.defaultFocusedComponent = comp;
  }

  public Component getDefaultFocusedComponent(){
    return this.defaultFocusedComponent;
  }

  public void setInputColor(Color color){
    System.out.println("input color set");
    StyleConstants.setForeground(inputStyle, color);
  }

  public void setDefaultColor(Color color){
    System.out.println("def color set");
    StyleConstants.setForeground(defaultStyle, color);
  }

  public void setTextBackgroundColor(Color color){
    inputField.setBackground(color);
    outputPane.setBackground(color);
  }

  private class ConsoleWriter
      extends Writer{

    /**
     * Print the specified part of argument char[] to console
     * @param chars char[]
     * @param offset int
     * @param len int
     */
    public void write(char[] chars, int offset, int len){
      Console.this.print(new String(chars, offset, len), defaultStyle);
    }

    public void flush(){} //Flushes automatically

    public void close(){} //implement later

  }

  private class ConsoleListener extends MouseAdapter
      implements ActionListener{

    public void actionPerformed(ActionEvent e){
      String s = inputField.getText();
      inputField.setText("");
      if(defaultFocusedComponent != null){
        inputField.setEnabled(false);
        defaultFocusedComponent.requestFocusInWindow();
      }
      if(!s.equals("")){
        //println(s, inputStyle);
        String response = parser.parse(s);
        if(response != null){
          println(response, inputStyle);
        }
      }
    }

    public void mouseClicked(MouseEvent e){
      enableTyping();
    }

  }

}

