package jwars.gui;

import java.util.List;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Transparency;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferStrategy;
import java.awt.image.VolatileImage;

import jwars.control.Control;
import jwars.control.UnitSelectionListener;
import jwars.control.ViewPort;
import jwars.gui.units.Sprite;
import jwars.model.Collider;
import jwars.model.DynamicSprite;
import jwars.model.Unit;
import jwars.model.WarModel;
import jwars.model.map.CollisionTile;
import jwars.model.map.CoordinateSystem;
import jwars.model.map.TerrainTile;

public class WarPanel
    extends Canvas implements ViewPort.ViewPortListener, UnitSelectionListener {

  //Framerate counter variables
  private int frameCount, fps;
  private long lastTime;

  //Viewport variables
  private int x = 0, y = 0;
  public final int WIDTH, HEIGHT;
  public final ViewPort VIEWPORT;

  private BufferStrategy bs;

  private Color markerColor = Color.GREEN;

  private ClippingRectangle terrainClip = new ClippingRectangle(WarModel.
      WORLD.TERRAIN_MAP);
  private ClippingRectangle unitClip = new ClippingRectangle(WarModel.
      WORLD.COLLISION_MAP);

  private final static boolean drawDebugGrids = false;

  private VolatileImage backBuffer;//

  //private DynamicSprite dynamicSprite = null;

  public WarPanel(int width, int height){
    this.setPreferredSize(new Dimension(width, height));
    this.WIDTH = width;
    this.HEIGHT = height;
    this.VIEWPORT = Control.VIEWPORT;
    VIEWPORT.setSize(width, height);
    VIEWPORT.addListener(this);
    VIEWPORT.addListener(terrainClip);
    VIEWPORT.addListener(unitClip);
    VIEWPORT.setLocation(0, 0); //Maybe change initialization value
    this.setIgnoreRepaint(true);
    this.setFocusable(false);
    Control.UNIT_SELECTION.addListener(this);
    WarMouseAdapter listener = new WarMouseAdapter();
    this.addMouseListener(listener);
    this.addMouseWheelListener(listener);
  }

  public void render(){
    multiBufferedRender();
  }

  private void multiBufferedRender(){
    if(bs == null){
      this.createBufferStrategy(2);
      this.bs = this.getBufferStrategy();
    }
    if(backBuffer == null){
      backBuffer = jwars.Util.GRAPH_CONFIG.createCompatibleVolatileImage(getWidth(), getHeight(), Transparency.OPAQUE);
    }
    Graphics gBackBuffer = null;
    do{
      gBackBuffer = backBuffer.createGraphics();
      _render(gBackBuffer);
      gBackBuffer.dispose();
    } while(backBuffer.contentsLost());

    //Use active rendering
    Graphics g = bs.getDrawGraphics();
    renderBuffered(g);
    renderFastMovingStuff(g);
    g.dispose();
    bs.show();
  }

  private void standardRender(){
    if(bs == null){
      this.createBufferStrategy(2);
      this.bs = this.getBufferStrategy();
    }
    Graphics gStrategy = bs.getDrawGraphics();
    _render(gStrategy);
    gStrategy.dispose();
    bs.show();
  }

  private void renderBuffered(Graphics g){
    g.drawImage(backBuffer, 0, 0, null);
  }

  private void _render(Graphics g){
    TerrainTile t;
    for(int i = terrainClip.startIndexX; i <= terrainClip.endIndexX; i++){
      for(int j = terrainClip.startIndexY; j <= terrainClip.endIndexY; j++){
        t = WarModel.WORLD.TERRAIN_MAP.getTile(i, j);
        boolean isDirty = t.isDirty();
        if(isDirty){
          t.draw(g, t.xPix-x, t.yPix-y);
          //t.setDirty(false);
        }
      }
    }
    Graphics2D g2d = (Graphics2D)g;

    CollisionTile tile;
    List<Collider> colliders;
    Sprite sprite;
    int x, y; //Drawing locations in pixels
    Unit markedUnit = Control.UNIT_SELECTION.markedUnit();

    AffineTransform transform = g2d.getTransform();
    for(int i = unitClip.startIndexX; i <= unitClip.endIndexX; i++){
      for(int j = unitClip.startIndexY; j <= unitClip.endIndexY; j++){
        tile = WarModel.WORLD.COLLISION_MAP.getTile(i, j);
        colliders = tile.colliders;
        if(drawDebugGrids){
          Color oldColor = g.getColor();
          g.setColor(Color.black);
          drawCollisionTileRepresentation(tile,
              WarModel.WORLD.COLLISION_MAP.transformToCoords(
                  WarModel.PIXEL_COORDS, i)-this.x,
            WarModel.WORLD.COLLISION_MAP.transformToCoords(
                  WarModel.PIXEL_COORDS, j)-this.y,
              g);
          g.setColor(oldColor);
        }
        for(Collider c : colliders){
          sprite = c.getSprite();
          if(sprite.isDirty()){
            x = WarModel.WORLD.MAIN_COORDS.transformToCoords(
                WarModel.PIXEL_COORDS, c.getX());
            y = WarModel.WORLD.MAIN_COORDS.transformToCoords(
                WarModel.PIXEL_COORDS, c.getY());
            sprite.draw(g2d, x - this.x, y - this.y);
            //sprite.setDirty(false);
            g2d.setTransform(transform);
            if(markedUnit != null && markedUnit.superiorOf(c)){// m.subordinateOf(Control.UNIT_SELECTION.markedUnit())){
              drawMarker(c, g, x, y);
            }
          }
        }


      }
    }
    for(int i = terrainClip.startIndexX; i <= terrainClip.endIndexX; i++){
      for(int j = terrainClip.startIndexY; j <= terrainClip.endIndexY; j++){
        t = WarModel.WORLD.TERRAIN_MAP.getTile(i, j);
        boolean isDirty = t.isDirty();
        if(isDirty){
          t.drawVegetation(g, t.xPix-this.x, t.yPix-this.y);
          t.setDirty(false);
        }
      }
    }
    //drawFrameCount(g);
    /*if(drawDebugGrids){
      drawDebugGrids(g);
    }*/
  }

  private void renderFastMovingStuff(Graphics g){
    drawFrameCount(g);
    drawDynamicSprites(g);
    //g.drawString("hello", jwars.Util.TIMER.getTime() % getWidth(), getHeight()-10);
  }

  private int mainToScreenX(int x){
    return WarModel.WORLD.MAIN_COORDS.transformToCoords(WarModel.PIXEL_COORDS, x) - this.x;
  }

  private int mainToScreenY(int y){
    return WarModel.WORLD.MAIN_COORDS.transformToCoords(WarModel.PIXEL_COORDS, y) - this.y;
  }

  private void drawDynamicSprites(Graphics g){
    int x, y;
    //DynamicSprite previous = null;
    Graphics2D g2d = (Graphics2D)g;
    for(DynamicSprite sprite = WarModel.WORLD.getFirstDynamicSprite(); sprite != null; sprite = sprite.next()){
      x = mainToScreenX(sprite.getX());//WarModel.WORLD.MAIN_COORDS.transformToCoords(WarModel.PIXEL_COORDS, sprite.getX())-this.x;
      y = mainToScreenY(sprite.getY());//this.y - WarModel.WORLD.MAIN_COORDS.transformToCoords(WarModel.PIXEL_COORDS, sprite.getX());
      //boolean stillAlive =
      sprite.draw(g2d, x, y, jwars.Util.TIMER.getTime());
      //if(!stillAlive){
      //  previous.next = sprite.next;
      //  sprite.next = null;
      //  sprite = previous;
        //continue;
      //} else {
      //  previous = sprite;
      //}
    }

  }

  public void drawCollisionTileRepresentation(CollisionTile tile, int x, int y, Graphics g){
    int sideLength = WarModel.WORLD.COLLISION_MAP.transformToCoords(WarModel.PIXEL_COORDS, 1);
    g.drawString(""+tile.colliders.size(), x+sideLength/2, y+sideLength/2);
    g.drawRect(x, y, sideLength, sideLength);
  }

  private void drawMarker(Collider c, Graphics g, int x, int y){
    g.setColor(markerColor);
    int radius = WarModel.WORLD.MAIN_COORDS.transformToCoords(
        WarModel.PIXEL_COORDS, c.getRadius());
    int centerX = x - this.x;
    int centerY = y - this.y;
    g.drawOval(centerX - radius, centerY - radius, 2 * radius,
               2 * radius);
  }

  /**
   * Calculate framerate and draw it in top left corner of screen
   * @param g Graphics
   */
  private void drawFrameCount(Graphics g){
    //Currently does not draw to screen. Prints to console.
    frameCount++;
    long time = System.currentTimeMillis();
    if(time - lastTime > 1000){
      fps = 1000 * frameCount / (int)(time - lastTime);
      frameCount = 0;
      lastTime = time;
      //System.out.println("FPS: " + fps);
    }
    g.setColor(Color.green);
    g.drawString("FPS: "+fps, 2,12);
  }

  public void reportLocationChange(int x, int y){
    this.x = x;
    this.y = y;

    for(int i = terrainClip.startIndexX; i <= terrainClip.endIndexX; i++){
      for(int j = terrainClip.startIndexY; j <= terrainClip.endIndexY; j++){
        WarModel.WORLD.TERRAIN_MAP.getTile(i, j).setDirty(true);
      }
    }

  }

  public void markUnit(Unit u){
    Control.UNIT_SELECTION.markUnit(u);
  }

  private void dirtifyTree(Unit u){
    u.dirtify();
    for(int i = 0; i < u.getSubUnitCount(); i++){
      u.getUnit(i).dirtify();
    }
  }

  /**
   * Select the unit closest to (x,y), if a unit exists which is close
   * enough.
   * @param x int
   * @param y int
   */
  public void select(int x, int y){
    //unmarkUnit();
    CollisionTile tile = WarModel.WORLD.COLLISION_MAP.getTileAt(x, y);
    double currentDist = Double.MAX_VALUE;
    Collider closest = null;
    double dist = currentDist;
    for(Collider c : tile.colliders){
      dist = Math.hypot(c.getX() - x, c.getY() - y);
      if(dist < currentDist && dist < c.getRadius()){
        closest = c;
        currentDist = dist;
      }
    }
    if(closest == null){
      markUnit(null);
    } else if(closest instanceof Unit){
      markUnit((Unit)closest); //May be null
    }
  }

  /**
   * Order the currently selected unit to move towards (x,y).
   * @param x int
   * @param y int
   */
  public void orderMove(MouseEvent me, int x, int y){
    Unit markedUnit = Control.UNIT_SELECTION.markedUnit();
    if(markedUnit != null){
      if(me.isShiftDown()){
        markedUnit.appendMove(x,y);
      }
      else if(me.isControlDown()){
        markedUnit.prependMove(x,y);
      } else {
        markedUnit.applyMove(x, y);
      }
    }
  }

  public void unitSelected(Unit newSelection, Unit lastSelection){
    if(newSelection != null){
      dirtifyTree(newSelection);
    }
    if(lastSelection != null){
      dirtifyTree(lastSelection);
    }
  }

  /**
   * Utility class used for taking out chunks of TileMaps
   */
  private class ClippingRectangle
      implements ViewPort.ViewPortListener{

    private CoordinateSystem coords;
    public int startIndexX, startIndexY, endIndexX, endIndexY;

    public ClippingRectangle(CoordinateSystem coords){
      this.coords = coords;
    }

    public void reportLocationChange(int x, int y){
      startIndexX = WarModel.PIXEL_COORDS.transformToCoords(coords, x);
      startIndexY = WarModel.PIXEL_COORDS.transformToCoords(coords, y);
      endIndexX = WarModel.PIXEL_COORDS.transformToCoords(coords, x + WIDTH);
      endIndexY = WarModel.PIXEL_COORDS.transformToCoords(coords, y + HEIGHT);
    }

  }

  private class WarMouseAdapter
      extends MouseAdapter implements MouseWheelListener {

    public void mousePressed(MouseEvent e){
      int x = WarModel.PIXEL_COORDS.
          transformToCoords(WarModel.WORLD.MAIN_COORDS,
                            e.getX() + WarPanel.this.x);
      int y = WarModel.PIXEL_COORDS.
          transformToCoords(WarModel.WORLD.MAIN_COORDS,
                            e.getY() + WarPanel.this.y);
      if(e.getButton() == MouseEvent.BUTTON1){
        if(e.isShiftDown()){
          Control.UNIT_SELECTION.markedUnit().attack(x,y, 0);
        } else {
          select(x, y);
          Unit markedUnit = Control.UNIT_SELECTION.markedUnit();
          if(e.getClickCount() > 1 && markedUnit != null){
            for(int i = 1; i < e.getClickCount(); i++){
              markedUnit = markedUnit.getSuperUnit();
            }
            markUnit(markedUnit);
          }
        }
      }
      else if(e.getButton() == MouseEvent.BUTTON3){
        orderMove(e, x, y);
      }
    }

    public void mouseWheelMoved(MouseWheelEvent e){
      Unit markedUnit = Control.UNIT_SELECTION.markedUnit();
      if(e.getWheelRotation() < 0 && markedUnit != null){
        markUnit(markedUnit.getSuperUnit());
      }
    }

  }

}
