package jwars.model;

import jwars.gui.units.Sprite;
import jwars.model.map.CollisionTile;
import jwars.model.map.CoordinateSystem;
import jwars.model.map.TerrainTile;
import jwars.model.map.TileMap;

public class CollisionDetector{

  //Used for collision detection between units
  private final TileMap<CollisionTile> map;
  //Used for influence of terrain on units
  private final TileMap<TerrainTile> terrain;
  private final CoordinateSystem mainCoords;

  private final TileMap.TileOperation terrainDirtifier =
      new TileMap.TileOperation<TerrainTile>(){

    public void operate(TerrainTile tile){
      tile.setDirty(true);
    }

  };

  //Registers units into overlapping collision tiles
  private final Registrator registrator = new Registrator();
  //Removes units from overlapping collision tiles
  private final Registrator deRegistrator = new DeRegistrator();

  private final TileChecker tileChecker = new TileChecker();

  public CollisionDetector(World world){
    this.mainCoords = world.MAIN_COORDS;
    this.map = world.COLLISION_MAP;
    this.terrain = world.TERRAIN_MAP;
  }

  /**
   * Checks whether the specified unit may be located at (x,y) in the main
   * coordinate system without colliding with other units. If collision is
   * not detected, the unit will immediately be placed at its new location and
   * surrounding terrain- and collision tiles will be considered dirty.
   * @param u unit to be moved
   * @param x desired x-coordinate of the unit in main coordinate system
   * @param y desired y-coordinate of the unit in main coordinate system
   * @return true if the unit has been successfully moved, false otherwise
   */
  public boolean move(Moveable m, int x, int y){
    boolean isMoveValid = tileChecker.check(m, x, y);
    if(isMoveValid){
      deRegistrator.operateOn(m);
      //Must be done before AND after move, otherwise the box must be expanded
      //properly
      dirtify(m);

      m.location.x = x;
      m.location.y = y;

      register(m);
      dirtify(m);
    }
    return isMoveValid;
  }

  public boolean register(Collider c){
    int[] coords = mainCoords.transformToCoords(map, getBounds(c));
    boolean legalPosition = true;
    legalPosition &= map.isLegalXCoordinate(coords[0]);
    legalPosition &= map.isLegalYCoordinate(coords[1]);
    legalPosition &= map.isLegalXCoordinate(coords[2]);
    legalPosition &= map.isLegalYCoordinate(coords[3]);
    if(legalPosition){
      registrator.operateOn(c);
      c.getCollisionProps().generalLocation = map.getTileAt(c.getX(), c.getY());
    }
    if(!legalPosition)
      System.out.println("Warning: Collider " + c +
          " cannot be registered since it is outside collision grid");
    return legalPosition;
    //c.generalLocation = map.getTileAt(m.location.x, m.location.y);
  }

  /**
   * NOTE: MUST TAKE CARE OF DIFFERENT SPRITE RADII
   * @param m Moveable
   */
  void dirtify(Collider c){
    //int radius = WarModel.PIXEL_COORDS.transformToCoords(
    //  mainCoords, m.getSprite().getRadius());
    terrain.performOp(
        terrainDirtifier,
        mainCoords.transformToCoords(terrain, getBounds(c.getSprite(), c.getX(), c.getY())));
    //Problem with radius of sprite compared to radius of unit!
  }

  /**
   * Returns an array containing the following ordered elements:
   * the smallest x-value covered by c
   * the smallest y-value covered by c
   * the largest x-value covered by c
   * the largest y-value covered by c
   *
   * @param c Collider
   * @return int[]
   */
  public int[] getBounds(Collider c){
    int x = c.getX();
    int y = c.getY();
    int r = c.getRadius();
    coordBuffer[0] = x - r;
    coordBuffer[1] = y - r;
    coordBuffer[2] = x + r;
    coordBuffer[3] = y + r;
    return coordBuffer;
  }

  public int[] getBounds(Sprite s, int x, int y){
    int r = WarModel.PIXEL_COORDS.transformToCoords(mainCoords, s.getRadius());
    coordBuffer[0] = x - r;
    coordBuffer[1] = y - r;
    coordBuffer[2] = x + r;
    coordBuffer[3] = y + r;
    return coordBuffer;
  }

  /*public void setDirtyRegion(Unit u){
    int x = u.getX();
    int y = u.getY();
    int startIndexX = MAIN_COORDS.transformToCoords(TERRAIN_MAP, x-u.RADIUS);
    int startIndexY = MAIN_COORDS.transformToCoords(TERRAIN_MAP, y-u.RADIUS);
    int endIndexX = MAIN_COORDS.transformToCoords(TERRAIN_MAP, x+u.RADIUS);
    int endIndexY = MAIN_COORDS.transformToCoords(TERRAIN_MAP, y+u.RADIUS);


    //TerrainTile start = TERRAIN_MAP.getTileAt(x-u.RADIUS, y-u.RADIUS);
    //TerrainTile end = TERRAIN_MAP.getTileAt(x+u.RADIUS, y+u.RADIUS);
    for(int i = startIndexX; i <= endIndexX; i++){
      for(int j = startIndexY; j <= endIndexY; j++){
        TERRAIN_MAP.getTile(i,j).setDirty(true);
      }
    }

     }*/

  /**
   * This variable serves as a temporary storage for bounding boxes. Its
   * contents may be changed from many different methods, so use only for
   * operations that terminate immediately without calling other methods that
   * need or modify the buffer uncontrollably.
   */
  private int[] coordBuffer = new int[4];

  /*private int[] coords(int x1, int y1, int x2, int y2){
    coordBuffer[0] = x1;
    coordBuffer[1] = y1;
    coordBuffer[2] = x2;
    coordBuffer[3] = y2;
    return coordBuffer;
     }*/

  private class TileChecker
      implements TileMap.TileOperation<CollisionTile>{

    private Collider currentCollider;
    private int x, y;
    private boolean moveLegal;

    /**
     * Check whether the specified Moveable can be at the specified location
     * @param m Moveable to be considered
     * @param x int
     * @param y int
     * @return boolean whether the Moveable fits
     */
    public boolean check(Collider c, int x, int y){
      moveLegal = true;
      this.x = x;
      this.y = y;
      currentCollider = c;
      map.performOp(this, mainCoords.transformToCoords(map, getBounds(c)));
      //Execution should be stopped for efficiency if collision
      //is found, however there is no way of doing this nicely right now
      return moveLegal;
    }

    /*public boolean probe(Collider c, int x, int y){
      moveLegal = true;
      this.x = x;
      this.y = y;
      currentCollider = c;
      map.performOp(this, mainCoords.transformToCoords(map, getBounds(c)));
      //Execution should be stopped for efficiency if collision
      //is found, however there is no way of doing this nicely right now
      return moveLegal;
    }*/

    /**
     * Calculate distance between the proposed new location (x,y) and
     * the coordinates of some other unit u.
     * @param u Unit
     * @return int
     */
    private int dist(int x1, int y1, int x2, int y2){
      return(int)Math.ceil(Math.hypot(x1 - x2, y1 - y2));
    }

    public Collider shootTo(Collider c, int x, int y){
      int r = c.getRadius();
      return null;
    }

    public void operate(CollisionTile tile){
      CollisionProperties currentProps = currentCollider.getCollisionProps();
      int currentR = currentCollider.getRadius(),
          otherR,
          otherX,
          otherY,
          distance;
      for(Collider other : tile.colliders){
        //Check whether the specified point is 'too close'
        otherX = other.getX();
        otherY = other.getY();
        otherR = other.getRadius();
        distance = dist(x, y, otherX, otherY);
        if(distance < currentR + otherR && currentCollider != other){
          //current unit is NOT deregistered so we have to be careful
          //that it doesn't 'collide with itself'
          if(currentProps.mayCollide(other.getCollisionProps())){
            moveLegal = false;
            currentCollider.collisionDetected(other);
          }
        }
        //Now check whether the units actually overlap, which they shouldn't
        /*if(currentCollider != other &&
           dist(currentCollider.getX(), currentCollider.getY(), otherX, otherY)
           < currentR + otherR){
          //current unit is NOT deregistered
          System.out.println("ERROR: Units overlap: " + currentCollider + " " +
                             other);
                 }*/
      }
    }

  }

  /**
   * This class is used for registering units across all collision tiles
   * within their bounding box.
   */
  private class Registrator
      implements TileMap.TileOperation<CollisionTile>{

    Collider current;

    //This counter is used as a test, counting members of collision tiles
    //It may be safely removed
    int testCounter = 0;

    public void operateOn(Collider c){
      testCounter = 0;
      current = c;
      map.performOp(this, mainCoords.transformToCoords(map, getBounds(c)));
    }

    public void operate(CollisionTile tile){
      if(tile.colliders.contains(current))
        System.out.println("UNIT NOT PROPERLY DEREGISTERED");
      tile.register(current);
      testCounter++;
    }

  };

  /**
   * This class does the same as the above, except it removes the unit.
   */
  private class DeRegistrator
      extends Registrator{

    public void operate(CollisionTile tile){
      testCounter++;
      tile.remove(current);
    }

  };

}
