package jwars.model;

import java.awt.Graphics2D;

import static jwars.Util.minusPiToPi;
import jwars.data.MoveableCategory;
import jwars.gui.units.Sprite;
import jwars.model.ai.impl.MoveableAI;
import jwars.model.weapons.WeaponModel;

/**
 * Moveables represent physical entities which can move around, such as
 * vehicles or infantry squads. Units
 */
public class Moveable extends Unit implements Collider {

  public final static double dt = WarModel.TIME_PER_FRAME/1000d;
  public final static double second = WarModel.second;
  public final static double meter = WarModel.meter;

  public final static double kilometer = 1000*meter;
  public final static double minute = 60*second;
  public final static double hour = 60*minute;

  private final static Sprite EMPTY_SPRITE = new EmptySprite();

  private final CollisionProperties collisionProps;

  public final MoveableCategory CAT;

  //The orientation of this Unit. Measured in radians.
  //protected double theta;
  //Current and maximum speed of this Unit. This will be replaced by a model
  //supporting different terrain types later.
  protected double speed;
  //protected double turnSpeed = 2*Math.PI;
  private Sprite sprite;
  private MoveableAI controller; //The AI of this Moveable with proper type
  //We should provide a proper identifier at some time.

  protected final WeaponModel weaponry;// = new WeaponModel();

  /**
   * Construct a unit
   * @param radius radius of this Unit, measured in units, for collision detection
   * @param sprite Sprite responsible for rendering this unit
   * @param x x-value of this Unit's location
   * @param y y-value of this Units location
   * @param theta angle of this Unit, measured in radians, with respect to the
   * x-axis
   * @param maxSpeed Maximum speed this Unit is capable of moving.
   */
  public Moveable(MoveableCategory cat,/*, String name, int radius, double maxSpeed,*/ boolean massive){
    super(cat.name, cat.radius);
    this.CAT = cat;
    controller = new MoveableAI(this);
    super.setAI(controller);
    setSprite(EMPTY_SPRITE);
    weaponry = new WeaponModel(controller, cat.weaponry);
    //this.maxSpeed = maxSpeed;
    this.sprite = sprite;
    this.collisionProps = new CollisionProperties(massive, true);
  }

  /**
   * Update the state of this Unit
   */
  public void update(){
    if(controller.isFiring()){
      double targetAngle = controller.getTargetAngle();
      boolean canFire = aimTowards(targetAngle) && weaponry.canFire();
      if(canFire){
        weaponry.fire(getX(), getY(), controller.getTargetX(), controller.getTargetY()); //Make this method more sensible
      }
    }
    if(controller.isMoving()){
      move();
    }
  }

  //private static double turnSpeed = 0.1;

  /**
   * Move this Unit according to destination and speed. This method should be
   * invoked only from the Unit move queue.
   */
  public void move(){
    //int deltaX = destination.x - location.x;
    //int deltaY = destination.y - location.y;
    //this.theta = minusPiToPi(this.theta);
    double newTheta = minusPiToPi(controller.getMoveAngle());
    double angleDifference = minusPiToPi(newTheta - this.theta);
    double maxRotation = CAT.turnRate * dt;
    if(Math.abs(angleDifference) > maxRotation){
      this.theta += maxRotation * Math.signum(angleDifference);
      this.dirtify();
      return;
    } else {
      this.theta = newTheta;
    }
    //if(deltaX != 0)
    //  theta = Math.atan( (double)deltaY /(double)(deltaX) );
    //else{
    //  theta = deltaY < 0 ? Math.PI*3/2 : Math.PI/2;
    //}
    speed = CAT.speed;
    //theta = deltaX < 0 ? theta+Math.PI : theta;
    int dx = (int)(speed*Math.cos(theta)*dt);
    int dy = (int)(speed*Math.sin(theta)*dt);
    int newX = location.x + dx;
    int newY = location.y + dy;
    //if(Math.hypot(newX-controller.destination.x, newY - BRAIN.targetY) < Math.hypot(dx, dy)){
    //  BRAIN.objectiveReached();
    //}
    boolean hasMoved = WarModel.WORLD.DETECTOR.move(this,newX,newY);
    if(hasMoved){
      this.theta = theta; //Only actually turn if the move was successful
      controller.closeEnough((int)(speed * dt));
      //controller2.updateLocation((int)(speed*dt));
    }
  }

  /**
   * Add this Moveable to the world and register it in the collision detector.
   * @param world World
   * @param x int
   * @param y int
   */
  public void addToWorld(World world, int x, int y){
    super.addToWorld(world, x, y);
    world.DETECTOR.register(this);
  }

  /**
   * This method will thrown an UnsupportedOperationException. It is considered
   * illegal to add sub Units to Moveables, being the lowest level of
   * organization.
   *
   * @param u
   */
  public void addUnit(Unit u){
    throw new UnsupportedOperationException("Moveables cannot have sub units");
  }

  public boolean aimTowards(double theta){
    return true;
  }

  /**
   * Retrieve the angle of this Moveable in the main coordinate system.
   * @return double the angle
   */
  public double getTheta(){
    return theta;
  }

  /**
   * Alert collision detector and sprite rendering mechanisms of this Moveable
   * having changed.
   */
  public void dirtify(){
    sprite.setDirty(true);
    WarModel.WORLD.DETECTOR.dirtify(this);
  }

  /**
   * Retrieve the Sprite associated with this Moveable.
   * @return Sprite
   */
  public Sprite getSprite(){
    return sprite;
  }

  /**
   * Set the Sprite of this Moveable
   * @param sprite the new Sprite
   */
  public void setSprite(Sprite sprite){
    this.sprite = sprite;
  }

  /**
   * Fills the argument int[] of length at least 4 with the
   * bounding box data of this Moveable.
   *
   * @param storage int[]
   * @return int[]
   */
  public int[] getBounds(int[] storage){
    storage[0] = location.x-radius;
    storage[1] = location.y-radius;
    storage[2] = location.x+radius;
    storage[3] = location.y+radius;
    return storage;
  }

  /*public CollisionProperties props(){
    return collisionProps;
  }*/

  public void collisionDetected(Collider other){
    controller.walkAround(other);
  }

  public CollisionProperties getCollisionProps(){
    return collisionProps;
  }

  /*public void changeCollisionProperties(boolean massive, boolean mobile){
    collisionProps = new CollisionProperties(massive, mobile);
  }*/

  /**
   * This is the default Sprite implementation which is invisible. Units that
   * are visible should provide their own Sprite implementations.
   */
  private static class EmptySprite extends Sprite{

    private EmptySprite(){
      super(0);
    }

    public void draw(Graphics2D g2d, int x, int y){
    }

  }

}
