package jwars.model;

import java.util.ArrayList;
import java.util.List;

import jwars.model.map.CollisionTile;
import jwars.model.map.CoordinateSystem;
import jwars.model.map.StructureFactory;
import jwars.model.map.TerrainTile;
import jwars.model.map.TileMap;

/**
 * This class contains the state of the world, i.e. collision- and terrain
 * tilemaps, and stores all units in the game sorted by formations.
 */
public class World{

  public final CoordinateSystem MAIN_COORDS;
  public final TileMap<CollisionTile> COLLISION_MAP;
  public final TileMap<TerrainTile> TERRAIN_MAP;

  public final CollisionDetector DETECTOR;

  private final Unit EVERYONE = Unit.createContainerUnit();
  private final List<UnitListener> listeners = new ArrayList<UnitListener>();

  public final double meter;
  public final double second;

  //private final EventQueue overwatchQueue = new EventQueue(10);
  private DynamicSprite dynSprite = null;

  public World(CoordinateSystem mainCoords,
               TileMap<CollisionTile> collisionMap,
               TileMap<TerrainTile> terrainMap,
               double meter,
               double second){
    this.meter = meter;
    this.second = second;
    this.MAIN_COORDS = mainCoords;
    this.COLLISION_MAP = collisionMap;
    this.TERRAIN_MAP = terrainMap;
    this.DETECTOR = new CollisionDetector(this);
    EVERYONE.addToWorld(this, 0, 0);
    int width = (int)(MAIN_COORDS.WIDTH/WarModel.meter);
    int height = (int)(MAIN_COORDS.HEIGHT/WarModel.meter);
    System.out.println("World initialization completed. World size: "+width+"m times "+height+"m.");
  }

  //Units will be enumerated according to these tree
  private static int bitsPerNestingLevel = 5;
  private static int maxUnits = (1<<bitsPerNestingLevel) - 1;
  private static int treeDepth = 6;
  //30 of the 32 bits of an int are used to enumerate units


  /**
   * Fetch the Unit which has the argument ID
   * @param id ID of Unit to be fetched
   * @return Unit Unit with argument ID
   */
  public Unit getUnit(int id){
    Unit u = EVERYONE;
    for(int i = treeDepth; i >= 0; --i){
      u = u.getUnit(((id >>> (i * bitsPerNestingLevel)) & maxUnits) - 1);
    }
    return u;
  }

  /**
   * Used to set the initial conditions not provided by constructor.
   * For example, add units, set their objectives, and any other
   * configurations.
   */
  public void init(){
    //DefaultArmies armies = new DefaultArmies(this);
    StructureFactory.buildDefaultStructure(this);
    EVERYONE.checkAllIds(this); //debug
    System.out.println("Default world setup loaded");
    //addTopLevelUnit(armies.GERMANY);
    //addTopLevelUnit(armies.SOVIET_UNION);
    //EVERYONE.addChildrenToWorld(this);
    //Moveable germanHQ = new TestUnit();
    //Moveable russianHQ = new TestUnit();
    //Moveable germanTank = new Tank();
    //Moveable russianTank = new Tank();
    /*
    int companySize = (int)(100*WarModel.meter);
    int platoonSize = (int)(20*WarModel.meter);

    Formation germans = new Formation(companySize);
    Formation russians = new Formation(companySize);

    List<Tank> germanTanks = new ArrayList<Tank>();
    List<Tank> russianTanks = new ArrayList<Tank>();

    List<InfantrySquad> germanInfantry = new ArrayList<InfantrySquad>();
    List<InfantrySquad> russianInfantry = new ArrayList<InfantrySquad>();

    for(int i = 0; i < 4; i++){
      germanTanks.add(new Tank());
      russianTanks.add(new Tank());
      germanInfantry.add(new InfantrySquad());
      germanInfantry.add(new InfantrySquad());
      russianInfantry.add(new InfantrySquad());
      russianInfantry.add(new InfantrySquad());
    }

    //int size = 100 * (int)Moveable.meter;

    Formation
        geTanks = new Formation(platoonSize),
        ruTanks = new Formation(platoonSize),
        geInf = new Formation(platoonSize),
        ruInf = new Formation(platoonSize);

    geTanks.addUnits(germanTanks);
    ruTanks.addUnits(russianTanks);
    geInf.addUnits(germanInfantry);
    ruInf.addUnits(russianInfantry);

    germans.addUnit(geTanks);
    germans.addUnit(geInf);
    russians.addUnit(ruTanks);
    russians.addUnit(ruInf);
    //germans.addUnit(germanHQ);
    //germans.addUnit(germanTank);
    //russians.addUnit(russianHQ);
    //russians.addUnit(russianTank);

    addTopLevelUnit(germans);
    addTopLevelUnit(russians);

    int germanPosX = (int)(Moveable.meter * 200);
    int germanPosY = (int)(Moveable.meter * 400);
    int russianPosX = (int)(Moveable.meter * 600);
    int russianPosY = (int)(Moveable.meter * 400);

    germans.addToWorld(this, germanPosX, germanPosY);
    russians.addToWorld(this, russianPosX, russianPosY);

    EVERYONE.checkAllIds(this);
    */

    //Unit u = new TestUnit(100000,100000);
    //Unit v = new TestUnit(120000,130000);
    //Unit t34 = new Tank(200000,200000);

    //this.addTopLevelUnit(t34);
    //this.addTopLevelUnit(v);
    //this.addTopLevelUnit(u);
  }

  void registerUnit(Unit unit){
    for(UnitListener listener : listeners){
      listener.unitNowOnBoard(unit);
    }
  }

  public void addDynamicSprite(DynamicSprite sprite){
    sprite.next = dynSprite;
    if(dynSprite != null)
      dynSprite.previous = sprite;
    dynSprite = sprite;
  }

  public DynamicSprite getFirstDynamicSprite(){
    return dynSprite;
  }

  void removeDynamicSprite(DynamicSprite sprite){
    if(sprite.previous == null){
      dynSprite = sprite.next;
    } else {
      sprite.previous.next = sprite.next;
      sprite.next.previous = sprite.previous;
      sprite.previous = null;
      //Don't set sprite.next to null, since that will prevent it from
      //returning the correct object throught a next() call
    }
  }

  public void addTopLevelUnit(Unit unit){
    EVERYONE.addUnit(unit);
    //if(unit instanceof Moveable){
    //  Moveable m = (Moveable)unit;
    //  m.BRAIN.stop();
    //  DETECTOR.register(m);
    //}
  }

  public Unit getEveryOne(){
    return EVERYONE;
  }

  public void update(){
    EVERYONE.update();
  }

  public void addUnitListener(UnitListener listener){
    listeners.add(listener);
  }

  public void removeUnitListener(UnitListener listener){
    listeners.remove(listener);
  }

  public int relativeX(double x){
    return (int)(x*MAIN_COORDS.WIDTH);
  }

  public int relativeY(double y){
    return (int)(y*MAIN_COORDS.HEIGHT);
  }

  public int countTotalUnits(){
    return countSubUnits(EVERYONE);
  }

  public int countSubUnits(Unit u){
    int currentTotal = 0;
    if(u.getSubUnitCount() == 0){
      currentTotal++;
    } else {
      for(int i = 0; i < u.getSubUnitCount(); i++){
        currentTotal += countSubUnits(u.getUnit(i));
      }
    }
    return currentTotal;
  }

  public int secondsToFrames(double seconds){
    return (int)(1000 * seconds / WarModel.TIME_PER_FRAME);
  }

  public double framesToSeconds(int frameCount){
    return (double)frameCount * WarModel.TIME_PER_FRAME * 1000;
  }

  public double toMetres(int dist){
    return dist / WarModel.meter;
  }

  public int metric(double meters){
    return (int)(meters * WarModel.meter);
  }

}
