package jwars.model.map.generator;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Transparency;
import java.awt.image.BufferedImage;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;

/**
 * <p>q w</p>
 *
 * <p>q w</p>
 *
 * <p>e r</p>
 *
 * <p>r t</p> y u i o
 */
public class FloatBuffer{

  public final int WIDTH, HEIGHT;
  private final float[] buffer;
  public final int SIZE;

  public FloatBuffer(int width, int height){
    this.WIDTH = width;
    this.HEIGHT = height;
    this.buffer = new float[width*height];
    this.SIZE = width * height;
  }

  public float get(int i, int j){
    return buffer[i * HEIGHT + j];
  }

  public void set(int i, int j, float value){
    buffer[i * HEIGHT + j] = value;
  }

  public float get(int index){
    return buffer[index];
  }

  public float[] getBuffer(){
    return buffer;
  }

  public float maxValue(){
    float max = Float.NEGATIVE_INFINITY;
    for(float f : buffer){
      if(f > max){
        max = f;
      }
    }
    return max;
  }

  public float minValue(){
    float min = Float.POSITIVE_INFINITY;
    for(float f : buffer){
      if(f < min){
        min = f;
      }
    }
    return min;
  }

  /**
   * Distributes the values in this FloatBuffer such that the smallest value
   * is min and the largest is max.
   *
   * This method guarantees that no element exceeds the boundaries.
   *
   * @param min float
   * @param max float
   */
  public void scaleToFitStrict(float min, float max){
    scaleToFit(min,max);
    for(int i = 0; i < SIZE; i++){
      if(buffer[i]<min){
        buffer[i] = min;
      }
      if(buffer[i]>max){
        buffer[i] = max;
      }
    }
  }

  /**
   * Distributes the values in this FloatBuffer such that the smallest value
   * is min and the largest is max.
   *
   * The extrema of this FloatBuffer may slightly exceed the boundaries
   * due to non-exact floating point precision.
   *
   * @param min float
   * @param max float
   */
  public void scaleToFit(float min, float max){
    float oldMax = maxValue();
    float oldMin = minValue();
    float factor = (max-min)/(oldMax-oldMin);
    //System.out.println("factor "+factor);
    float displacement = min - factor * oldMin;
    scale(factor);
    add(displacement);
    //System.out.println("min "+minValue());
    //System.out.println("max "+maxValue());
  }

  private void add(float a){
    for(int i = 0; i < SIZE; i++){
      buffer[i] += a;
    }
  }

  private void scale(float s){
    for(int i = 0; i < SIZE; i++){
      buffer[i] *= s;
    }
  }

  public void cutOff(float minValue){
    add(-minValue);
    for(int i = 0; i < buffer.length; i++){
      if(buffer[i] < 0){
        buffer[i] = 0;
      }
    }
  }


  public FloatBuffer crop(int x0, int y0, int width, int height){
    FloatBuffer newBuffer = new FloatBuffer(width, height);
    for(int i = x0; i < x0+width; i++){
      for(int j = y0; j < y0 + height; j++){
        newBuffer.set(i-x0, j-y0, get(i,j));
      }
    }
    return newBuffer;
  }

  /**
   * Traverses all non-boundary elements of this FloatBuffer and sets them to
   * the average of their neighbours.
   */
  public void smoothify(){
    for(int i = 1; i < WIDTH-1; i++){
      for(int j = 1; j < HEIGHT-1; j++){
        float sum = get(i-1,j) + get(i+1, j) + get(i, j-1) + get(i, j+1);
        set(i, j, sum * 0.25f);
      }
    }
  }

  /**
   * For debugging
   */
  public void display(){
    BufferedImage im = new BufferedImage(WIDTH, HEIGHT, Transparency.OPAQUE);
    Graphics g = im.getGraphics();
    for(int i=0; i < WIDTH; i++){
      for(int j = 0; j < HEIGHT; j++){
        float height = get(i, j);
        Color c;
        float max = Math.max( Math.abs(maxValue()), Math.abs(minValue()) );

        try{
          if(height > 0){
            c = new Color(height / max, 0, 0);
          } else{
            c = new Color(0, 0, -height / max);
          }
          g.setColor(c);
          g.drawLine(i, j, i, j);
        }
        catch(Exception ex){
          ex.printStackTrace();
          System.out.println("height "+height+ "maxh "+max);
          System.out.println("i j "+i+" "+j);
          System.exit(0);
        }

      }
    }

    JFrame frame = new JFrame();
    frame.add(new JLabel(new ImageIcon(im)));
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.pack();
    frame.setVisible(true);
  }

  public static void main(String[] args){
    HeightGenerator h = new HeightGenerator(0L, 7);
    FloatBuffer f = h.generateHeights();
    f.scaleToFit(0,4);
    System.out.println(""+f.maxValue()+" "+f.minValue());
  }

}
